//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "misivi.h"

#if ENABLE_RENDERING

#include "TransGl/glDraw.h"
#include "TransGl/viewer.h"
#include "TransProg/r32Vector.h"
#include "TransProg/surface.h"

#define SNAPSHOT_RECORDING 0

#if SNAPSHOT_RECORDING
  bool snapshotRecording=false;
  int snapshotCount=0;
  int snapshotW=-1;
  int snapshotH=-1;
  uint8_t *snapshotStorage=(uint8_t *)0;
#endif

#define BUFFER_CAPACITY (4*1024*1024)

static real32_t cpuColors[4*ASSUMED_MAX_CPU_COUNT];

static real32_t bacteriumColors[4*ASSUMED_MAX_BACTERIUM_TYPE_COUNT];

typedef struct _ThreadRenderingState
{
CACHE_ALIGNED
real64_t xSeen,ySeen,zSeen;
struct
  {
  R32Vector xyz;
  R32Vector color;
  R32Vector radius;
  } bacterium[ASSUMED_MAX_BACTERIUM_TYPE_COUNT];
} ThreadRenderingState;

typedef struct _ThreadRenderingData
{
ThreadRenderingState state[2];
} ThreadRenderingData;

typedef struct _RenderingData
{
GlShape lines;
GlShape sideShape[6]; // W,E,S,N,D,U
int sideTexture[6]; // W,E,S,N,D,U
int texBuffer;
bool needsSideShapeRebuild;
bool showCpuColors;
int rendered;
real32_t radiusFactor;
GlPointSphere pointSphere;
Viewer viewer;
intptr_t seenCount;
real64_t xSeen,ySeen,zSeen;
ThreadRenderingData trd[ASSUMED_MAX_CPU_COUNT];
} RenderingData;

static RenderingData rd;

int
Rendering_sideTextureBuffer(void)
{
return rd.texBuffer;
}

INLINE
int
nextPowerOfTwo(int i)
{
--i;
i|=i>>1;
i|=i>>2;
i|=i>>4;
i|=i>>8;
i|=i>>16;
return ++i;
}

static
void
rebuildSideShapes(void)
{
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,substrate_sampling);
IMPORT_INT_CONSTANT(cst,substrate_plane_count);
IMPORT_R32_CONSTANT(cst,substrate_size);
IMPORT_R32_CONSTANT(cst,substrate_step);
const Substrate * RESTRICT sub=&sd.substrate;
const int * RESTRICT sides=sub->sides;
const real32_t * RESTRICT planeDepths=sub->planeDepths;
const real32_t * RESTRICT planeSteps=sub->planeSteps;
const real32_t xyTexStep=1.0f/nextPowerOfTwo(cst_substrate_sampling);
const real32_t zTexStep=1.0f/nextPowerOfTwo(cst_substrate_plane_count);
for(int side=0;side<6;++side) // W,E,S,N,D,U
  {
  int pAxis=-1,hAxis=-1,vAxis=-1;
  int firstH=-1,lastH=-1;
  int firstV=-1,lastV=-1;
  real32_t h0=-1.0f,dh=-1.0f;
  real32_t v0=-1.0f,dv=-1.0f;
  real32_t p0=-1.0f,pN=-1.0f;
  bool ccw=false;
  switch(side)
    {
    case WEST_SIDE:
    case EAST_SIDE:
      {
      pAxis=0; hAxis=1; vAxis=2;
      firstH=sides[SOUTH_SIDE];     lastH=sides[NORTH_SIDE];
      firstV=sides[UP_SIDE];        lastV=sides[DOWN_SIDE];
      h0=-0.5f*cst_substrate_size;  dh=cst_substrate_step;
      v0=0.0f;                      // dv depends upon v
      if(side==WEST_SIDE)
        {
        p0=-0.5f*cst_substrate_size+sides[side]*cst_substrate_step;
        pN=-1.0f;
        ccw=true;
        }
      else // EAST_SIDE
        {
        p0=-0.5f*cst_substrate_size+(sides[side]+1)*cst_substrate_step;
        pN=+1.0f;
        ccw=false;
        }
      break;
      }
    case SOUTH_SIDE:
    case NORTH_SIDE:
      {
      pAxis=1; hAxis=0; vAxis=2;
      firstH=sides[WEST_SIDE];      lastH=sides[EAST_SIDE];
      firstV=sides[UP_SIDE];        lastV=sides[DOWN_SIDE];
      h0=-0.5f*cst_substrate_size;  dh=cst_substrate_step;
      v0=0.0f;                      // dv depends upon v
      if(side==SOUTH_SIDE)
        {
        p0=-0.5f*cst_substrate_size+sides[side]*cst_substrate_step;
        pN=-1.0f;
        ccw=false;
        }
      else // NORTH_SIDE
        {
        p0=-0.5f*cst_substrate_size+(sides[side]+1)*cst_substrate_step;
        pN=+1.0f;
        ccw=true;
        }
      break;
      }
    case DOWN_SIDE:
    case UP_SIDE:
      {
      pAxis=2; hAxis=0; vAxis=1;
      firstH=sides[WEST_SIDE];      lastH=sides[EAST_SIDE];
      firstV=sides[SOUTH_SIDE];     lastV=sides[NORTH_SIDE];
      h0=-0.5f*cst_substrate_size;  dh=cst_substrate_step;
      v0=-0.5f*cst_substrate_size;  dv=cst_substrate_step;
      if(side==DOWN_SIDE)
        {
        p0=0.0f-(planeDepths[sides[side]]+planeSteps[sides[side]]);
        pN=-1.0f;
        ccw=false;
        }
      else // UP_SIDE
        {
        p0=0.0f-planeDepths[sides[side]];
        pN=+1.0f;
        ccw=true;
        }
      break;
      }
    }
  const int h_end=2+lastH-firstH;
  const int h_step=h_end-1;
  const int hCount=2;
  const int v_end=2+lastV-firstV;
  const int v_step=vAxis==2 ? 1 : v_end-1;
  const int vCount=vAxis==2 ? v_end : 2;
  const real32_t dht=xyTexStep;
  const real32_t dvt=(vAxis==2 ? zTexStep : xyTexStep);
  const int indexCount=(hCount-1)*(vCount-1)*2*3;
  const int vertexCount=hCount*vCount;
  int *indices=Memory_alloc(indexCount,int);
  real32_t *vertices=Memory_alloc(vertexCount*3,real32_t);
  real32_t *normals=Memory_alloc(vertexCount*3,real32_t);
  real32_t *texCoords=Memory_alloc(vertexCount*2,real32_t);
  int iIdx=0;
  int vIdx=0;
  for(int v=0;v<v_end;v+=v_step)
    {
    const int vCell=firstV+v;
    const real32_t vCoord=v0+(vAxis==2 ? -planeDepths[vCell] : vCell*dv);
    const real32_t vTexCoord=vCell*dvt;
    for(int h=0;h<h_end;h+=h_step)
      {
      const int hCell=firstH+h;
      const real32_t hCoord=h0+hCell*dh;
      const real32_t hTexCoord=hCell*dht;
      vertices[3*vIdx+hAxis]=hCoord;
      vertices[3*vIdx+vAxis]=vCoord;
      vertices[3*vIdx+pAxis]=p0;
      normals[3*vIdx+hAxis]=0.0f;
      normals[3*vIdx+vAxis]=0.0f;
      normals[3*vIdx+pAxis]=pN;
      texCoords[2*vIdx+0]=hTexCoord;
      texCoords[2*vIdx+1]=vTexCoord;
      ++vIdx;
      if(h&&v)
        {
        const int hp=h/h_step;
        const int vp=v/v_step;
        const int p0=(vp-1)*hCount+(hp-1);
        const int p1=(vp-1)*hCount+(hp-0);
        const int p2=(vp-0)*hCount+(hp-0);
        const int p3=(vp-0)*hCount+(hp-1);
        indices[iIdx++]=p0;
        indices[iIdx++]=ccw ? p1 : p3;
        indices[iIdx++]=p2;
        indices[iIdx++]=p2;
        indices[iIdx++]=ccw ? p3 : p1;
        indices[iIdx++]=p0;
        }
      }
    }
  GlShape *shape=&rd.sideShape[side];
  shape->indexCount=indexCount;
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,GlShape_indexBuffer(shape));
  glBufferSubData(GL_ELEMENT_ARRAY_BUFFER,
                  0,indexCount*sizeof(int),indices);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);
  shape->vertexCount=vertexCount;
  glBindBuffer(GL_ARRAY_BUFFER,GlShape_vertexBuffer(shape));
  glBufferSubData(GL_ARRAY_BUFFER,
                  0,vertexCount*3*sizeof(real32_t),vertices);
  glBindBuffer(GL_ARRAY_BUFFER,GlShape_normalBuffer(shape));
  glBufferSubData(GL_ARRAY_BUFFER,
                  0,vertexCount*3*sizeof(real32_t),normals);
  glBindBuffer(GL_ARRAY_BUFFER,GlShape_texCoordBuffer(shape));
  glBufferSubData(GL_ARRAY_BUFFER,
                  0,vertexCount*2*sizeof(real32_t),texCoords);
  glBindBuffer(GL_ARRAY_BUFFER,0);
  Memory_free(indices);
  Memory_free(vertices);
  Memory_free(normals);
  Memory_free(texCoords);
  }
rd.needsSideShapeRebuild=false;
}

static
void
viewerCB(Viewer *v)
{
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,substrate_sampling);
IMPORT_R32_CONSTANT(cst,substrate_size);
IMPORT_INT_CONSTANT(cst,substrate_plane_count);
IMPORT_INT_CONSTANT(cst,bacterium_type_count);
IMPORT_R32_CONSTANT(cst,diffusivity_plot_amount);
Substrate * RESTRICT sub=&sd.substrate;
int * RESTRICT sides=sub->sides;
GlView *view=Viewer_view(v);
const GfxState *state=GlView_state(view);
real64_t d=rd.seenCount ? 1.0/rd.seenCount : 0.0;
Viewer_setRefDistanceFromPoint(v,d*rd.xSeen,d*rd.ySeen,d*rd.zSeen);
switch(GfxState_event(state))
  {
  case GFX_KEY_EVT:
    {
    if(GfxState_pressed(state))
      {
      if(!strcmp(GfxState_key(state),"F11"))
        {
        Perspective_setStereoEyeSeparation(Viewer_perspective(v),
                    Perspective_stereoEyeSeparation(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"F12"))
        {
        Perspective_setStereoViewDistance(Viewer_perspective(v),
                    Perspective_stereoViewDistance(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"Home"))
        {
        if(GfxState_modifiers(state)&GFX_SHIFT_MDF)
          { sd.needsSubstrateReset=true; }
        else
          {
          SpatialBasis_init(Viewer_viewPoint(&rd.viewer));
          SpatialBasis_yaw(Viewer_viewPoint(&rd.viewer),M_PI_2);
          SpatialBasis_pitch(Viewer_viewPoint(&rd.viewer),M_PI_2);
          SpatialBasis_translate(Viewer_viewPoint(&rd.viewer),
                                 -0.9*cst_substrate_size,0.0,0.0);
          }
        }
      else if(!strcmp(GfxState_key(state),"\t"))
        {
        const int cycle=cst_diffusivity_plot_amount ? SUBSTANCE_COUNT+1
                                                    : SUBSTANCE_COUNT;
        rd.rendered=(rd.rendered+1)%cycle;
        const char *name;
        switch(rd.rendered)
          {
          case SUBSTANCE_COUNT+0: name="diffusivity"; break;
          default: name=sd.substrate.substances[rd.rendered].name;
          }
        Platform_msg("~~~~~~~~~~~~~~~~ rendering %s ~~~~~~~~~~~~~~~~\n",
                     name);
        }
      else if(!strcmp(GfxState_key(state),"Left"))
        {
        if(sides[WEST_SIDE]>0)
          { rd.needsSideShapeRebuild=true; --sides[WEST_SIDE]; }
        else if(sides[EAST_SIDE]>sides[WEST_SIDE])
          { rd.needsSideShapeRebuild=true; --sides[EAST_SIDE]; }
        }
      else if(!strcmp(GfxState_key(state),"Right"))
        {
        if(sides[EAST_SIDE]<cst_substrate_sampling-1)
          { rd.needsSideShapeRebuild=true; ++sides[EAST_SIDE]; }
        else if(sides[WEST_SIDE]<sides[EAST_SIDE])
          { rd.needsSideShapeRebuild=true; ++sides[WEST_SIDE]; }
        }
      else if(!strcmp(GfxState_key(state),"Down"))
        {
        if(sides[SOUTH_SIDE]>0)
          { rd.needsSideShapeRebuild=true; --sides[SOUTH_SIDE]; }
        else if(sides[NORTH_SIDE]>sides[SOUTH_SIDE])
          { rd.needsSideShapeRebuild=true; --sides[NORTH_SIDE]; }
        }
      else if(!strcmp(GfxState_key(state),"Up"))
        {
        if(sides[NORTH_SIDE]<cst_substrate_sampling-1)
          { rd.needsSideShapeRebuild=true; ++sides[NORTH_SIDE]; }
        else if(sides[SOUTH_SIDE]<sides[NORTH_SIDE])
          { rd.needsSideShapeRebuild=true; ++sides[SOUTH_SIDE]; }
        }
      else if(!strcmp(GfxState_key(state),"Prior"))
        {
        if(sides[UP_SIDE]>0)
          { rd.needsSideShapeRebuild=true; --sides[UP_SIDE]; }
        else if(sides[DOWN_SIDE]>sides[UP_SIDE])
          { rd.needsSideShapeRebuild=true; --sides[DOWN_SIDE]; }
        }
      else if(!strcmp(GfxState_key(state),"Next"))
        {
        if(sides[DOWN_SIDE]<cst_substrate_plane_count-1)
          { rd.needsSideShapeRebuild=true; ++sides[DOWN_SIDE]; }
        else if(sides[UP_SIDE]<sides[DOWN_SIDE])
          { rd.needsSideShapeRebuild=true; ++sides[UP_SIDE]; }
        }
      else if(!strcmp(GfxState_key(state),"1"))
        {
        rd.needsSideShapeRebuild=true;
        sides[WEST_SIDE]=0;
        sides[EAST_SIDE]=cst_substrate_sampling-1;
        sides[SOUTH_SIDE]=0;
        sides[NORTH_SIDE]=cst_substrate_sampling-1;
        sides[DOWN_SIDE]=cst_substrate_plane_count-1;
        sides[UP_SIDE]=0;
        }
      else if(!strcmp(GfxState_key(state),"2"))
        {
        rd.needsSideShapeRebuild=true;
        sides[WEST_SIDE]=0;
        sides[EAST_SIDE]=cst_substrate_sampling-1;
        sides[SOUTH_SIDE]=cst_substrate_sampling>>1;
        sides[NORTH_SIDE]=cst_substrate_sampling-1;
        sides[DOWN_SIDE]=cst_substrate_plane_count-1;
        sides[UP_SIDE]=0;
        }
      else if(!strcmp(GfxState_key(state),"4"))
        {
        rd.needsSideShapeRebuild=true;
        sides[WEST_SIDE]=cst_substrate_sampling>>1;
        sides[EAST_SIDE]=cst_substrate_sampling-1;
        sides[SOUTH_SIDE]=cst_substrate_sampling>>1;
        sides[NORTH_SIDE]=cst_substrate_sampling-1;
        sides[DOWN_SIDE]=cst_substrate_plane_count-1;
        sides[UP_SIDE]=0;
        }
      else if(!strcmp(GfxState_key(state),"c"))
        {
        rd.showCpuColors=!rd.showCpuColors;
        }
      else if(!strcmp(GfxState_key(state)," "))
        {
        sd.pause=!sd.pause;
        if(sd.pause) Platform_msg("==== PAUSE ====\n");
        }
#if SNAPSHOT_RECORDING
      else if(!strcmp(GfxState_key(state),"r"))
        {
        snapshotRecording=!snapshotRecording;
        if(snapshotRecording) Platform_msg("==== RECORDING ====\n");
        else Platform_msg("==== NOT RECORDING ====\n");
        }
#endif
      else if(!strcmp(GfxState_key(state),"\x1B"))
        {
        sd.mustQuit=true;
        }
      }
    break;
    }
  case GFX_BUTTON_EVT:
    {
    if(GfxState_pressed(state))
      {
      switch(GfxState_button(state))
        {
        case GFX_UP_BTN:
          {
          rd.radiusFactor=fminf(100.0f,rd.radiusFactor/0.8f);
          break;
          }
        case GFX_DOWN_BTN:
          {
          rd.radiusFactor=fmaxf(1.0f,rd.radiusFactor*0.8f);
          break;
          }
        default: break;
        }
      }
    break;
    }
  case GFX_CLOSE_EVT:
    {
    sd.mustQuit=true;
    break;
    }
  case GFX_DRAW_EVT:
    {
    if(rd.needsSideShapeRebuild)
      {
      rebuildSideShapes();
      }
    rd.xSeen=rd.ySeen=rd.zSeen=0.0;
    rd.seenCount=0;
    const real32_t *projMtx=Viewer_projectionMatrix(v);
    const real32_t *viewMtx=Viewer_viewMatrix(v);
    const real32_t headLight[3]={0.0f,0.0f,0.0f};
    const real32_t shininess=20.0f;
    const real32_t ambient=0.2f;
    const real32_t substrateSpecular=0.0f;
    const real32_t substrateRgba[4]={0.9f,0.9f,0.9f,1.0f};
    const real32_t bacteriaSpecular=0.5f;
    const real32_t *fog=(const real32_t *)0;
    GlShape_draw(&rd.lines,
                 substrateRgba,ambient,substrateSpecular,shininess,
                 headLight,fog,0,projMtx,viewMtx);
    for(int side=0;side<6;++side) // W,E,S,N,D,U
      {
      GlShape_draw(&rd.sideShape[side],
                   substrateRgba,ambient,substrateSpecular,shininess,
                   headLight,fog,rd.sideTexture[side],projMtx,viewMtx);
      }
    const int fromState2=Scheduler_fromState2(&sd.sched);
    for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
      {
      const ThreadRenderingState * RESTRICT trs=&rd.trd[i].state[fromState2];
      rd.xSeen+=trs->xSeen;
      rd.ySeen+=trs->ySeen;
      rd.zSeen+=trs->zSeen;
      for(int b=0;b<cst_bacterium_type_count;++b)
        {
        const R32Vector * RESTRICT xyz=&trs->bacterium[b].xyz;
        const R32Vector * RESTRICT color=&trs->bacterium[b].color;
        const R32Vector * RESTRICT radius=&trs->bacterium[b].radius;
        const intptr_t seenCount=R32Vector_size(radius);
        if(seenCount)
          {
          rd.seenCount+=seenCount;
          const int vertexBuffer=GlPointSphere_vertexBuffer(&rd.pointSphere);
          const int colorBuffer=GlPointSphere_colorBuffer(&rd.pointSphere);
          const int radiusBuffer=GlPointSphere_radiusBuffer(&rd.pointSphere);
          const real32_t *vertexData=R32Vector_data(xyz);
          const real32_t *colorData=R32Vector_data(color);
          const real32_t *radiusData=R32Vector_data(radius);
          intptr_t offset=0;
          while(offset<seenCount)
            {
            intptr_t amount=seenCount-offset;
            if(amount>BUFFER_CAPACITY) amount=BUFFER_CAPACITY;
            glBindBuffer(GL_ARRAY_BUFFER,vertexBuffer);
            OPENGL_CHECK;
            glBufferSubData(GL_ARRAY_BUFFER,
                            0,amount*3*sizeof(real32_t),
                            vertexData+3*offset);
            OPENGL_CHECK;
            glBindBuffer(GL_ARRAY_BUFFER,colorBuffer);
            OPENGL_CHECK;
            glBufferSubData(GL_ARRAY_BUFFER,
                            0,amount*3*sizeof(real32_t),
                            colorData+3*offset);
            OPENGL_CHECK;
            glBindBuffer(GL_ARRAY_BUFFER,radiusBuffer);
            OPENGL_CHECK;
            glBufferSubData(GL_ARRAY_BUFFER,
                            0,amount*sizeof(real32_t),
                            radiusData+offset);
            OPENGL_CHECK;
            glBindBuffer(GL_ARRAY_BUFFER,0);
            OPENGL_CHECK;
            rd.pointSphere.vertexCount=amount;
            GlPointSphere_draw(&rd.pointSphere,
                               rd.showCpuColors ? cpuColors+4*i
                                                : bacteriumColors+4*b,
                               ambient,bacteriaSpecular,shininess,
                               rd.radiusFactor,headLight,fog,projMtx,viewMtx);
            offset+=amount;
            }
          }
        }
      }
    break;
    }
  default:
    {
    break;
    }
  }
}

void
Rendering_showBacterium(int cpuCtx,
                        int type,
                        real32_t x,
                        real32_t y,
                        real32_t radius,
                        int level,
                        real32_t growthAmount)
{
const int toState2=Scheduler_toState2(&sd.sched);
const Viewer *viewer=&rd.viewer;
if(GfxState_visible(Viewer_state(viewer)))
  {
  const ConstantRecord * RESTRICT cst=ConstantRecord_get();
  IMPORT_R32_CONSTANT(cst,radius_min);
  IMPORT_R32_CONSTANT(cst,radius_range);
  const real32_t diameter_mid=cst_radius_min+0.5f*cst_radius_range;
  const real32_t z=(level+0.5f)*diameter_mid;
#if 1
  real64_t lx=x,ly=y,lz=z;
  SpatialBasis_globalToLocalPosition(Viewer_viewPoint(viewer),&lx,&ly,&lz);
  if(Perspective_locateVertex(Viewer_perspective(viewer),lx,ly,lz)
     ==LOC_MIDDLE)
#endif
    {
    ThreadRenderingState * RESTRICT trs=&rd.trd[cpuCtx].state[toState2];
    trs->xSeen+=x;
    trs->ySeen+=y;
    trs->zSeen+=z;
    R32Vector * RESTRICT xyzVector=&trs->bacterium[type].xyz;
    R32Vector * RESTRICT colorVector=&trs->bacterium[type].color;
    R32Vector * RESTRICT radiusVector=&trs->bacterium[type].radius;
    R32Vector_push(xyzVector,x);
    R32Vector_push(xyzVector,y);
    R32Vector_push(xyzVector,z);
    const real32_t factor=clamp32(1.0f-growthAmount,0.0f,1.0f);
    const real32_t shade=1.0f-0.5f*factor; //*factor;
    R32Vector_push(colorVector,shade);
    R32Vector_push(colorVector,shade);
    R32Vector_push(colorVector,shade);
    R32Vector_push(radiusVector,radius);
    }
  }
}

void
Rendering_activate(int cpuCtx,
                   void *data)
{
(void)data;
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,bacterium_type_count);
IMPORT_INT_CONSTANT(cst,substrate_sampling);
IMPORT_INT_CONSTANT(cst,substrate_plane_count);
bool needsGfxSync=false;
const int toState2=Scheduler_toState2(&sd.sched);
if(sd.needsRedraw[toState2])
  {
  ThreadRenderingState * RESTRICT trs=&rd.trd[cpuCtx].state[toState2];
  trs->xSeen=0.0;
  trs->ySeen=0.0;
  trs->zSeen=0.0;
  for(int b=0;b<cst_bacterium_type_count;++b)
    {
    R32Vector_clear(&trs->bacterium[b].xyz);
    R32Vector_clear(&trs->bacterium[b].color);
    R32Vector_clear(&trs->bacterium[b].radius);
    }
  }
if(!cpuCtx)
  {
  if(sd.needsRedraw[toState2])
    {
    for(int side=0;side<6;++side) // W,E,S,N,D,U
      {
      const int sideWidth=cst_substrate_sampling;
      const int sideHeight=side<DOWN_SIDE ? cst_substrate_plane_count
                                          : cst_substrate_sampling;
      Substrate_updateTextureBuffer(side,sideWidth,sideHeight,rd.rendered);
      glBindTexture(GL_TEXTURE_2D,rd.sideTexture[side]);
      OPENGL_CHECK;
      glBindBuffer(GL_PIXEL_UNPACK_BUFFER,rd.texBuffer);
      OPENGL_CHECK;
      glTexSubImage2D(GL_TEXTURE_2D,0,0,0,
                      sideWidth,sideHeight,
                      GL_BGRA,GL_UNSIGNED_INT_8_8_8_8_REV,
                      (const void *)0);
      OPENGL_CHECK;
      }
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER,0);
    OPENGL_CHECK;
    }
  const int fromState2=Scheduler_fromState2(&sd.sched);
  if(sd.needsRedraw[fromState2])
    {
    GlView_postRedraw(Viewer_view(&rd.viewer));
    GlView_processEvents(0);
    needsGfxSync=true;
    }
  }
ActivationTile_simulate(cpuCtx,(void *)0);
if(needsGfxSync)
  {
  GlView_sync();
#if SNAPSHOT_RECORDING
  if(snapshotRecording)
    {
    char buffer[0x20];
    sprintf(buffer,"snapshot_%6.6d.tga",snapshotCount++);
    GlView *view=Viewer_view(&rd.viewer);
    const GfxState *state=GlView_state(view);
    int w=GfxState_width(state);
    int h=GfxState_height(state);
    if((w!=snapshotW)||(h!=snapshotH))
      {
      snapshotW=w;
      snapshotH=h;
      snapshotStorage=Memory_reAlloc(snapshotStorage,3*w*h,uint8_t);
      }
    GlView_makeCurrentGlTarget(view);
    GlDraw_snapshot(0,0,w,h,snapshotStorage);
    Gfx_saveTgaImage(buffer,w,h,3,snapshotStorage);
    }
#endif
  }
}

void
Rendering_init(void)
{
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_R32_CONSTANT(cst,substrate_size);
IMPORT_R32_CONSTANT(cst,substrate_thickness);
IMPORT_INT_CONSTANT(cst,substrate_plane_count);
IMPORT_INT_CONSTANT(cst,substrate_sampling);
IMPORT_INT_CONSTANT(cst,bacterium_type_count);
IMPORT_R32_CONSTANT(cst,radius_min);
GlView_init();
GlDraw_init();
//-- Substrate --
const real32_t westCoord=-0.5f*cst_substrate_size;
const real32_t eastCoord=+0.5f*cst_substrate_size;
const real32_t southCoord=-0.5f*cst_substrate_size;
const real32_t northCoord=+0.5f*cst_substrate_size;
const real32_t downCoord=0.0f-cst_substrate_thickness;
const real32_t upCoord=0.0f;
const real32_t vertices[8*3]={ westCoord, southCoord, downCoord,
                               eastCoord, southCoord, downCoord,
                               eastCoord, northCoord, downCoord,
                               westCoord, northCoord, downCoord,
                               westCoord, southCoord, upCoord,
                               eastCoord, southCoord, upCoord,
                               eastCoord, northCoord, upCoord,
                               westCoord, northCoord, upCoord };
const int indices[12*2]={ 0,1, 1,2, 2,3, 3,0,
                          0,4, 1,5, 2,6, 3,7,
                          4,5, 5,6, 6,7, 7,4 };
GlShape_init(&rd.lines,0,GL_LINES,12*2,indices,8,vertices,
             (const real32_t *)0,(const real32_t *)0,(const real32_t *)0);
int texBufferSize=0;
for(int side=0;side<6;++side) // W,E,S,N,D,U
  {
  const int sideWidth=cst_substrate_sampling;
  const int sideHeight=side<DOWN_SIDE ? cst_substrate_plane_count
                                      : cst_substrate_sampling;
  if(sideWidth>texBufferSize)
    { texBufferSize=sideWidth; }
  if(sideHeight>texBufferSize)
    { texBufferSize=sideHeight; }
  const intptr_t indexCount=sideWidth*sideHeight*2*3;
  const intptr_t vertexCount=(sideWidth+1)*(sideHeight+1);
  GlShape_init(&rd.sideShape[side],
               GLSHAPE_HAS_NORMALS|GLSHAPE_HAS_TEXCOORDS,
               GL_TRIANGLES,indexCount,(const int *)0,vertexCount,
               (const real32_t *)0,(const real32_t *)0,
               (const real32_t *)0,(const real32_t *)0);
  glGenTextures(1,(GLuint *)&rd.sideTexture[side]);
  OPENGL_CHECK;
  glBindTexture(GL_TEXTURE_2D,rd.sideTexture[side]);
  OPENGL_CHECK;
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  OPENGL_CHECK;
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  OPENGL_CHECK;
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  OPENGL_CHECK;
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
  OPENGL_CHECK;
  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA8,
               nextPowerOfTwo(sideWidth),nextPowerOfTwo(sideHeight),
               0,GL_BGRA,GL_UNSIGNED_INT_8_8_8_8_REV,(const void *)0);
  OPENGL_CHECK;
  }
texBufferSize=nextPowerOfTwo(texBufferSize);
glGenBuffers(1,(GLuint *)&rd.texBuffer);
OPENGL_CHECK;
glBindBuffer(GL_COPY_WRITE_BUFFER,rd.texBuffer);
OPENGL_CHECK;
glBufferData(GL_COPY_WRITE_BUFFER,texBufferSize*texBufferSize*4,
             NULL,GL_DYNAMIC_DRAW);
OPENGL_CHECK;
rd.needsSideShapeRebuild=true;
rd.showCpuColors=false;
rd.rendered=0;
//-- bacteria --
rd.radiusFactor=1.0f;
for(int b=0;b<cst_bacterium_type_count;++b)
  {
  const BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+b;
  IMPORT_R32_CONSTANT(bcst,display_red);
  IMPORT_R32_CONSTANT(bcst,display_green);
  IMPORT_R32_CONSTANT(bcst,display_blue);
  real32_t * RESTRICT c=bacteriumColors+4*b;
  c[0]=bcst_display_red;
  c[1]=bcst_display_green;
  c[2]=bcst_display_blue;
  c[3]=1.0f;
  }
int cpuCount=Platform_cpuCtxCount();
int stepCount;
for(stepCount=0;stepCount*stepCount*stepCount<cpuCount;++stepCount) {}
real32_t colorStep=1.0f/(1+stepCount);
for(int blue=0;blue<stepCount;++blue)
  {
  for(int green=0;green<stepCount;++green)
    {
    for(int red=0;red<stepCount;++red)
      {
      int i=red+(green+blue*stepCount)*stepCount;
      if(i<cpuCount)
        {
        real32_t * RESTRICT c=cpuColors+4*i;
        c[0]=(1+red)*colorStep;
        c[1]=(1+green)*colorStep;
        c[2]=(1+blue)*colorStep;
        c[3]=1.0f;
        }
      }
    }
  }
GlPointSphere_init(&rd.pointSphere,
                   GLPOINTSPHERE_HAS_COLORS|GLPOINTSPHERE_HAS_RADIUSES,
                   BUFFER_CAPACITY,(const real32_t *)0,
                   (const real32_t *)0,(const real32_t *)0);
rd.seenCount=0;
rd.xSeen=0.0f;
rd.ySeen=0.0f;
rd.zSeen=0.0f;
for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
  {
  for(int s=0;s<2;++s)
    {
    ThreadRenderingState * RESTRICT trs=&rd.trd[i].state[s];
    trs->xSeen=0.0;
    trs->ySeen=0.0;
    trs->zSeen=0.0;
    for(int b=0;b<cst_bacterium_type_count;++b)
      {
      R32Vector_init(&trs->bacterium[i].xyz);
      R32Vector_init(&trs->bacterium[i].color);
      R32Vector_init(&trs->bacterium[i].radius);
      }
    }
  }
//-- Viewer --
#if SNAPSHOT_RECORDING
  int w=1024+14;
  int h=768+29;
#else
  int w=800;
  int h=600;
#endif
int x=GlView_screenWidth()-w;
int y=0;
Viewer_init(&rd.viewer,x,y,w,h,false,viewerCB,(void *)0);
GlView_setTitle(Viewer_view(&rd.viewer),"TransProg_Misivi");
SpatialBasis_yaw(Viewer_viewPoint(&rd.viewer),M_PI_2);
SpatialBasis_pitch(Viewer_viewPoint(&rd.viewer),M_PI_2);
SpatialBasis_translate(Viewer_viewPoint(&rd.viewer),
                       -0.9*cst_substrate_size,0.0,0.0);
real64_t farD=4.0*fmaxf(cst_substrate_size,cst_substrate_thickness);
real64_t nearD=cst_radius_min==FLT_MAX ? farD*0.01
                                       : 4.0*cst_radius_min;
Perspective_setFarDistance(Viewer_perspective(&rd.viewer),farD);
Perspective_setNearDistance(Viewer_perspective(&rd.viewer),nearD);
Perspective_setStereoEyeSeparation(Viewer_perspective(&rd.viewer),5.0*nearD);
Perspective_setStereoViewDistance(Viewer_perspective(&rd.viewer),
                                  0.5*(nearD+farD));
}

#endif // ENABLE_RENDERING

//----------------------------------------------------------------------------
