//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "misivi.h"

#define CLIP_NEGATIVE_VALUES 0

/***

Diffusion equation:

  C(x,y,z) --> concentration at ( x ; y ; z )
  D --> diffusion factor
  P(x,y,z) --> production rate at ( x ; y ; z )

  dC(x,y,z)/dt = P(x,y,z) + D * ( d2C(x,y,z)/dx2 +
                                  d2C(x,y,z)/dy2 +
                                  d2C(x,y,z)/dz2 )

General case, surrounding cell centers are at x+{hE|hW}:

                 C(x+hE/2,y,z) - C(x+hW/2,y,z)
  dC(x,y,z)/dx = -----------------------------
                 (hE-hW)/2

                   C(x+hE,y,z) + C(x+hW,y,z) - 2*C(x+(hE+hW)/2,y,z)
  d2C(x,y,z)/dx2 = ------------------------------------------------
                   ( (hE-hW)/2 )^2

  (similar along y and z)

Special case, regular grid with hE=-hW=hX:

                   C(x+hX,y,z) + C(x-hX,y,z) - 2*C(x,y,z)
  d2C(x,y,z)/dx2 = --------------------------------------
                   hX^2

  (similar along y and z)

Special case, regular grid with hE=-hW=hN=-hS=hD=-hU=h:

                            D
  dC(x,y,z)/dt = P(x,y,z) + --- * ( C(x+h,y,z) + C(x-h,y,z) +
                            h^2     C(x,y+h,z) + C(x,y-h,z) +
                                    C(x,y,z+h) + C(x,y,z-h) - 6*C(x,y,z) )

Backward-Euler method:

  C0(x,y,z) --> concentration at previous time step
  C1(x,y,z) --> concentration at next time step

  C1(x,y,z)-C0(x,y,z)
  ------------------- = dC1(x,y,z)/dt
          dt

  C1(x,y,z)*(1+6*s) - s*( C1(x+h,y,z) + ... + C1(x,y,z-h) ) =
  dt*P(x,y,z) + C0(x,y,z)

  with s=D*dt/h^2

  Resolution with Jacobi method:

    Solve matrix/vector form [A].|C1 = |B   with [A]=[ 1+6*s  -s ... ]
                                                 |B= dt*P + C0

                    1
    next_C(x,y,z) = ----- * ( B(x,y,z) + s * ( prev_C(x+h,y,z) + ... +
                    1+6*s                      prev_C(x,y,z-h) ) )

Crank–Nicolson method:

  C0(x,y,z) --> concentration at previous time step
  C1(x,y,z) --> concentration at next time step

  C1(x,y,z)-C0(x,y,z)
  ------------------- = 1/2 * ( dC1(x,y,z)/dt + dC0(x,y,z)/dt )
          dt

  C1(x,y,z)*(1+3*s) - (s/2)*( C1(x+h,y,z) + ... + C1(x,y,z-h) ) =
  dt*P(x,y,z) + C0(x,y,z)*(1-3*s) + (s/2)*( C0(x+h,y,z) + ... + C0(x,y,z-h) )

  with s=D*dt/h^2

  Resolution with Jacobi method:

    Solve matrix/vector form [A].|C1 = |B   with [A]=[ 1+3*s  -s/2 ... ]
                                                 |B= dt*P + C0*(1-3*s) + ...

                    1
    next_C(x,y,z) = ----- * ( B(x,y,z) + s/2 * ( prev_C(x+h,y,z) + ... +
                    1+3*s                        prev_C(x,y,z-h) ) )

Geometric progression of vertical cell size:

  hi --> size of cell i
  a*hi --> size of cell i+1
  hi/a --> size of cell i-1

  hD =  hi/2 * (1 + a)
  hU = -hi/2 * (1 + 1/a)

                 C(x,y,z+hD/2) - C(x,y,z+hU/2)
  dC(x,y,z)/dz = -----------------------------
                 (hD-hW)/2

                   C(x,y,z+hD) - C(x,y,z)   C(x,y,z) - C(x,y,z+hU)
  d2C(x,y,z)/dz2 = ---------------------- - ----------------------
                   hD*(hD-hU)/2             -hU*(hD-hU)/2

                   C(x,y,z+hD) - C(x,y,z)     C(x,y,z) - C(x,y,z+hU)
  d2C(x,y,z)/dz2 = ------------------------ - --------------------------
                   (hi^2/8)*(2+a+1/a)*(1+a)   (hi^2/8)*(2+a+1/a)*(1+1/a)


  changes in Backward-Euler, Crank-Nicolson and Jacobi:

    1+6*s    becomes    1+s*(4 + up_coeff + down_coeff)
    1+3*s    becomes    1+(s/2)*(4 + up_coeff + down_coeff)
    1-3*s    becomes    1-(s/2)*(4 + up_coeff + down_coeff)
    every C(x,y,z+hD)    is multiplied by    down_coeff
    every C(x,y,z+hU)    is multiplied by    up_coeff

    with up_coeff=8*h^2/(hi^2*(2+a+1/a)*(1+1/a))
         down_coeff=8*h^2/(hi^2*(2+a+1/a)*(1+a))
***/

static
void
Substrate_buildProgram(CudaProgram *out_program,
                       const CudaDevice *dev,
                       const char *name,
                       const Text *sourceCode)
{
CudaProgram_init(out_program,dev,name,
                 "--use_fast_math",Text_get(sourceCode),true);
const char *log=CudaProgram_buildLog(out_program);
if(log)
  {
  Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
               "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n%s"
               "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
               "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n",
               log);
  }
if(CudaProgram_buildFailure(out_program))
  {
  DROP_DEAD("gpu code compilation failure");
  }
}

void
Substrate_init(void)
{
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_R32_CONSTANT(cst,substrate_step);
IMPORT_R32_CONSTANT(cst,substrate_size);
IMPORT_R32_CONSTANT(cst,substrate_plane_progression);
IMPORT_INT_CONSTANT(cst,substrate_sampling);
IMPORT_INT_CONSTANT(cst,substrate_plane_count);
IMPORT_R32_CONSTANT(cst,diffusivity_plot_width);
IMPORT_R32_CONSTANT(cst,diffusivity_plot_amount);
IMPORT_R32_CONSTANT(cst,diffusivity_plot_value);
IMPORT_R32_CONSTANT(cst,initial_glucose_concentration);
IMPORT_R32_CONSTANT(cst,glucose_diffusivity);
IMPORT_R32_CONSTANT(cst,display_glucose_min);
IMPORT_R32_CONSTANT(cst,display_glucose_max);
IMPORT_R32_CONSTANT(cst,initial_lactate_concentration);
IMPORT_R32_CONSTANT(cst,lactate_diffusivity);
IMPORT_R32_CONSTANT(cst,display_lactate_min);
IMPORT_R32_CONSTANT(cst,display_lactate_max);
Substrate * RESTRICT sub=&sd.substrate;
sub->planeDepths=Memory_alloc(cst_substrate_plane_count+1,real32_t);
sub->planeSteps=Memory_alloc(cst_substrate_plane_count+1,real32_t);
sub->planeDepths[0]=0.0f;
sub->planeSteps[0]=cst_substrate_step;
for(int p=1;p<=cst_substrate_plane_count;++p)
  {
  sub->planeDepths[p]=sub->planeDepths[p-1]+sub->planeSteps[p-1];
  sub->planeSteps[p]=sub->planeSteps[p-1]*cst_substrate_plane_progression;
  }
const int topCellCount=cst_substrate_sampling*cst_substrate_sampling;
for(int s=0;s<2;++s)
  {
  sub->masses[s]=Memory_initAlloc(topCellCount,real32_t);
  }
Cuda_init(ENABLE_RENDERING);
const CudaDevice *dev=(const CudaDevice *)0;
#if ENABLE_RENDERING
  dev=Cuda_glDevice();
#endif
if(!dev)
  {
  int deviceCount;
  dev=Cuda_devices(&deviceCount); // use the first device
  }
CudaStream_init(&sub->compStream,dev);
CudaStream_init(&sub->sendStream,dev);
CudaStream_init(&sub->recvStream,dev);
const int topSize=topCellCount*(int)sizeof(real32_t);
const int bufSize=topSize*cst_substrate_plane_count;
CudaBuffer_init(&sub->gpu_ref,dev,bufSize);
CudaBuffer_init(&sub->gpu_tmp,dev,bufSize);
CudaBuffer_init(&sub->gpu_diff,dev,topSize);
real32_t *diff_array=(real32_t *)Memory_allocVoid(topSize);
const int cellCount=cst_substrate_sampling*cst_substrate_sampling;
for(int id=0;id<cellCount;++id)
  {
  diff_array[id]=1.0f;
  }
const int plotSize=(int)roundf(cst_diffusivity_plot_width/cst_substrate_step);
const real32_t plotSurface=plotSize*cst_substrate_step*
                           plotSize*cst_substrate_step;
const real32_t surface=cst_substrate_size*cst_substrate_size;
const int plotCount=(int)roundf(cst_diffusivity_plot_amount*surface/
                                plotSurface);
const int cpuCtx=Platform_cpuCtx();
for(int p=0;p<plotCount;++p)
  {
  int px,py;
  for(int attempt=0;attempt<10000;++attempt)
    {
    px=Random_int(cpuCtx,cst_substrate_sampling);
    py=Random_int(cpuCtx,cst_substrate_sampling);
    const int px1=(px+plotSize-1)%cst_substrate_sampling;
    const int py1=(py+plotSize-1)%cst_substrate_sampling;
    if((diff_array[py*cst_substrate_sampling+px]==1.0f)&&
       (diff_array[py*cst_substrate_sampling+px1]==1.0f)&&
       (diff_array[py1*cst_substrate_sampling+px]==1.0f)&&
       (diff_array[py1*cst_substrate_sampling+px1]==1.0f))
      {
      break; // no overlap, keep (px,py)
      }
    }
  for(int y=0;y<plotSize;++y)
    {
    const int iy=(py+y)%cst_substrate_sampling;
    for(int x=0;x<plotSize;++x)
      {
      const int ix=(px+x)%cst_substrate_sampling;
      const int id=iy*cst_substrate_sampling+ix;
      diff_array[id]=cst_diffusivity_plot_value;
      }
    }
  }
CudaBuffer_send(&sub->gpu_diff,0,topSize,diff_array,&sub->sendStream);
CudaStream_hostSync(&sub->sendStream);
Memory_free(diff_array);
sub->substances[GLUCOSE].name="glucose";
sub->substances[GLUCOSE].initValue=cst_initial_glucose_concentration;
sub->substances[GLUCOSE].diffusivity=cst_glucose_diffusivity;
sub->substances[GLUCOSE].minDisplay=cst_display_glucose_min;
sub->substances[GLUCOSE].maxDisplay=cst_display_glucose_max;
sub->substances[LACTATE].name="lactate";
sub->substances[LACTATE].initValue=cst_initial_lactate_concentration;
sub->substances[LACTATE].diffusivity=cst_lactate_diffusivity;
sub->substances[LACTATE].minDisplay=cst_display_lactate_min;
sub->substances[LACTATE].maxDisplay=cst_display_lactate_max;
for(int i=0;i<3;++i)
  {
  for(int j=0;j<SUBSTANCE_COUNT;++j)
    {
    Substance * RESTRICT s=&sub->substances[j];
    CudaBuffer_init(&s->gpu_c[i],dev,bufSize);
    CudaBuffer_init(&s->gpu_dc[i],dev,topSize);
    CudaLockedMem_init(&s->mem_c[i],false,topSize);
    CudaLockedMem_init(&s->mem_dc[i],false,topSize);
    s->top_c[i]=(real32_t *)CudaLockedMem_data(&s->mem_c[i]);
    s->top_dc[i]=(real32_t *)CudaLockedMem_data(&s->mem_dc[i]);
    }
  CudaMarker_init(&sub->compSync[i],dev);
  CudaMarker_init(&sub->sendSync[i],dev);
  CudaMarker_init(&sub->recvSync[i],dev);
  }
const CudaDeviceProperties *prop=CudaDevice_properties(dev);
const int warpSize=prop->warp_size;
const int maxWorkGroupSize=prop->max_block_dim_x;
int sz=warpSize;
while((sz<<1)<=maxWorkGroupSize) sz<<=1; // power-of-two
sub->groupSize=sz<warpSize ? warpSize : sz;
sub->groupCount=8*prop->multiprocessor_count;
Text header,src;
Text_init(&header);
Text_init(&src);
Text_appendStr(&header,cst->constantSourceCode);
Text_appendStr(&header,
  "                                                                    \n"
  " typedef float real32_t;                                            \n"
  "                                                                    \n"
  " #define DT             (cst_dt/cst_substrate_dt_divisor)           \n"
  " #define H2             (cst_substrate_step*cst_substrate_step)     \n"
  " #if cst_substrate_crank_nicolson                                   \n"
  " # define COEFF_B       (0.5f*DT/H2)                                \n"
  " #else // backward Euler                                            \n"
  " # define COEFF_B       (DT/H2)                                     \n"
  " #endif                                                             \n"
  " #define ONE_PLUS_A     (1.0f+cst_substrate_plane_progression)      \n"
  " #define ONE_PLUS_DIV_A (1.0f+1.0f/cst_substrate_plane_progression) \n"
  " #define UP_DOWN_NUMER  (8.0f*H2/(ONE_PLUS_A+ONE_PLUS_DIV_A))       \n"
  " #define DOWN_COEFF     (UP_DOWN_NUMER/ONE_PLUS_A)                  \n"
  " #define UP_COEFF       (UP_DOWN_NUMER/ONE_PLUS_DIV_A)              \n"
  " #define DIV_H2_PROGRESSION                                       \\\n"
  "         (1.0f/(cst_substrate_plane_progression*                  \\\n"
  "                cst_substrate_plane_progression))                   \n"
  "                                                                    \n"
  " #define X_COUNT        cst_substrate_sampling                      \n"
  " #define Y_COUNT        cst_substrate_sampling                      \n"
  " #define Z_COUNT        cst_substrate_plane_count                   \n"
  " #define X_LAST         (X_COUNT-1)                                 \n"
  " #define Y_LAST         (Y_COUNT-1)                                 \n"
  " #define Z_LAST         (Z_COUNT-1)                                 \n"
  " #define XY_COUNT       (X_COUNT*Y_COUNT)                           \n"
  " #define XYZ_COUNT      (XY_COUNT*Z_COUNT)                          \n"
  "                                                                    \n");
Text_append(&header,
  " #define CLIP_NEGATIVE_VALUES %d                       \n"
  " #define WARP_SIZE            %d                       \n"
  " #define GROUP_SIZE           %d                       \n"
  " #define GROUP_COUNT          %d                       \n"
  " #define GRID_SIZE            (GROUP_COUNT*GROUP_SIZE) \n"
  "                                                       \n",
  CLIP_NEGATIVE_VALUES,
  warpSize,
  sub->groupSize,
  sub->groupCount);
Text_appendText(&src,&header);
Text_appendStr(&src,
  " extern \"C\" __global__                       \n"
  " void                                          \n"
  " rstPrg(real32_t * __restrict__ value_array,   \n"
  "        const int count,                       \n"
  "        const real32_t initValue)              \n"
  " {                                             \n"
  " for(int id=GROUP_SIZE*blockIdx.x+threadIdx.x; \n"
  "     id<count;id+=GRID_SIZE)                   \n"
  "   {                                           \n"
  "   value_array[id]=initValue;                  \n"
  "   }                                           \n"
  " }                                             \n");
Substrate_buildProgram(&sub->rstPrg,dev,"rstPrg",&src);
Text_clear(&src);
Text_appendText(&src,&header);
Text_appendStr(&src,
  " extern \"C\" __global__                                       \n"
  " void                                                          \n"
  " refPrg(const real32_t diff,                                   \n"
  "        const real32_t * __restrict__ diff_array,              \n"
  "        const real32_t * __restrict__ value_array,             \n"
  "        const real32_t * __restrict__ delta_array,             \n"
  "        real32_t * __restrict__ ref_array)                     \n"
  " #if cst_substrate_crank_nicolson                              \n"
  " {                                                             \n"
  " for(int id=GROUP_SIZE*blockIdx.x+threadIdx.x;                 \n"
  "     id<XY_COUNT;id+=GRID_SIZE)                                \n"
  "   {                                                           \n"
  "   const int x=id%X_COUNT;                                     \n"
  "   const int y=id/X_COUNT;                                     \n"
  " #if cst_substrate_force_borders                               \n"
  "   const bool unchanged=!(x&&(X_LAST-x)&&y&&(Y_LAST-y));       \n"
  " #else                                                         \n"
  "   const bool unchanged=false;                                 \n"
  " #endif                                                        \n"
  "   int cId=id;                                                 \n"
  "   int wId=((x+X_LAST)%X_COUNT)+y*X_COUNT;                     \n"
  "   int eId=((x+1)%X_COUNT)+y*X_COUNT;                          \n"
  "   int sId=x+((y+Y_LAST)%Y_COUNT)*X_COUNT;                     \n"
  "   int nId=x+((y+1)%Y_COUNT)*X_COUNT;                          \n"
  "   const real32_t duCoeffB=COEFF_B*diff*diff_array[cId];       \n"
  "   const real32_t wCoeffB=0.5f*(duCoeffB+                      \n"
  "                                COEFF_B*diff*diff_array[wId]); \n"
  "   const real32_t eCoeffB=0.5f*(duCoeffB+                      \n"
  "                                COEFF_B*diff*diff_array[eId]); \n"
  "   const real32_t sCoeffB=0.5f*(duCoeffB+                      \n"
  "                                COEFF_B*diff*diff_array[sId]); \n"
  "   const real32_t nCoeffB=0.5f*(duCoeffB+                      \n"
  "                                COEFF_B*diff*diff_array[nId]); \n"
  "   const real32_t coeffA_part=1.0f-                            \n"
  "                              wCoeffB-eCoeffB-sCoeffB-nCoeffB; \n"
  "   real32_t center=value_array[cId];                           \n"
  "   real32_t down=center;                                       \n"
  "   real32_t downCoeff=DOWN_COEFF/H2;                           \n"
  "   real32_t upCoeff=UP_COEFF/H2;                               \n"
  "   real32_t delta=delta_array[cId];                            \n");
Text_appendStr(&src,
  "   for(const int bottomId=cId+(XY_COUNT*Z_LAST);               \n"
  "       cId<=bottomId;cId+=XY_COUNT)                            \n"
  "     {                                                         \n"
  "     const real32_t up=center;                                 \n"
  "     center=down;                                              \n"
  "     down=cId==bottomId ? center : value_array[cId+XY_COUNT];  \n"
  "     real32_t result;                                          \n"
  "     if(unchanged) { result=center;}                           \n"
  "     else                                                      \n"
  "       {                                                       \n"
  "       const real32_t west=value_array[wId];                   \n"
  "       const real32_t east=value_array[eId];                   \n"
  "       const real32_t south=value_array[sId];                  \n"
  "       const real32_t north=value_array[nId];                  \n"
  "       const real32_t coeffA=coeffA_part-                      \n"
  "                             duCoeffB*(downCoeff+upCoeff);     \n"
  "       result=delta+coeffA*center                              \n"
  "                   +wCoeffB*west+eCoeffB*east                  \n"
  "                   +sCoeffB*south+nCoeffB*north                \n"
  "                   +duCoeffB*(downCoeff*down+upCoeff*up);      \n"
  " #if CLIP_NEGATIVE_VALUES                                      \n"
  "       result=fmaxf(0.0f,result);                              \n"
  " #endif                                                        \n"
  "       }                                                       \n"
  "     ref_array[cId]=result;                                    \n"
  "     delta=0.0f;                                               \n"
  "     downCoeff*=DIV_H2_PROGRESSION;                            \n"
  "     upCoeff*=DIV_H2_PROGRESSION;                              \n"
  "     wId+=XY_COUNT; eId+=XY_COUNT;                             \n"
  "     sId+=XY_COUNT; nId+=XY_COUNT;                             \n"
  "     }                                                         \n"
  "   }                                                           \n"
  " }                                                             \n"
  " #else // backward Euler                                       \n"
  " {                                                             \n"
  " (void)diff;(void)diff_array;                                  \n"
  " for(int id=GROUP_SIZE*blockIdx.x+threadIdx.x;                 \n"
  "     id<XYZ_COUNT;id+=GRID_SIZE)                               \n"
  "   {                                                           \n"
  "   const real32_t value=value_array[id];                       \n"
  "   const real32_t delta=id<XY_COUNT ? delta_array[id] : 0.0f;  \n"
  "   real32_t result=delta+value;                                \n"
  " #if CLIP_NEGATIVE_VALUES                                      \n"
  "   result=fmaxf(0.0f,result);                                  \n"
  " #endif                                                        \n"
  "   ref_array[id]=result;                                       \n"
  "   }                                                           \n"
  " }                                                             \n"
  " #endif                                                        \n");
Substrate_buildProgram(&sub->refPrg,dev,"refPrg",&src);
Text_clear(&src);
Text_appendText(&src,&header);
Text_appendStr(&src,
  " extern \"C\" __global__                                          \n"
  " void                                                             \n"
  " iterPrg(const real32_t diff,                                     \n"
  "         const real32_t * __restrict__ diff_array,                \n"
  "         const real32_t * __restrict__ ref_array,                 \n"
  "         const real32_t * __restrict__ from_array,                \n"
  "         real32_t * __restrict__ to_array)                        \n"
  " {                                                                \n"
  " for(int id=GROUP_SIZE*blockIdx.x+threadIdx.x;                    \n"
  "     id<XY_COUNT;id+=GRID_SIZE)                                   \n"
  "   {                                                              \n"
  "   const int x=id%X_COUNT;                                        \n"
  "   const int y=id/X_COUNT;                                        \n"
  " #if cst_substrate_force_borders                                  \n"
  "   const bool unchanged=!(x&&(X_LAST-x)&&y&&(Y_LAST-y));          \n"
  " #else                                                            \n"
  "   const bool unchanged=false;                                    \n"
  " #endif                                                           \n"
  "   int cId=id;                                                    \n"
  "   int wId=((x+X_LAST)%X_COUNT)+y*X_COUNT;                        \n"
  "   int eId=((x+1)%X_COUNT)+y*X_COUNT;                             \n"
  "   int sId=x+((y+Y_LAST)%Y_COUNT)*X_COUNT;                        \n"
  "   int nId=x+((y+1)%Y_COUNT)*X_COUNT;                             \n"
  "   const real32_t duCoeffB=COEFF_B*diff*diff_array[cId];          \n"
  "   const real32_t wCoeffB=0.5f*(duCoeffB+                         \n"
  "                                COEFF_B*diff*diff_array[wId]);    \n"
  "   const real32_t eCoeffB=0.5f*(duCoeffB+                         \n"
  "                                COEFF_B*diff*diff_array[eId]);    \n"
  "   const real32_t sCoeffB=0.5f*(duCoeffB+                         \n"
  "                                COEFF_B*diff*diff_array[sId]);    \n"
  "   const real32_t nCoeffB=0.5f*(duCoeffB+                         \n"
  "                                COEFF_B*diff*diff_array[nId]);    \n"
  "   const real32_t coeffA_part=1.0f+                               \n"
  "                              wCoeffB+eCoeffB+sCoeffB+nCoeffB;    \n"
  "   real32_t center=from_array[cId];                               \n"
  "   real32_t down=center;                                          \n"
  "   real32_t downCoeff=DOWN_COEFF/H2;                              \n"
  "   real32_t upCoeff=UP_COEFF/H2;                                  \n");
Text_appendStr(&src,
  "   for(const int bottomId=cId+(XY_COUNT*Z_LAST);                  \n"
  "       cId<=bottomId;cId+=XY_COUNT)                               \n"
  "     {                                                            \n"
  "     const real32_t up=center;                                    \n"
  "     center=down;                                                 \n"
  "     down=cId==bottomId ? center : from_array[cId+XY_COUNT];      \n"
  "     real32_t result;                                             \n"
  "     if(unchanged) { result=center;}                              \n"
  "     else                                                         \n"
  "       {                                                          \n"
  "       const real32_t ref=ref_array[cId];                         \n"
  "       const real32_t west=from_array[wId];                       \n"
  "       const real32_t east=from_array[eId];                       \n"
  "       const real32_t south=from_array[sId];                      \n"
  "       const real32_t north=from_array[nId];                      \n"
  "       const real32_t coeffA=1.0f/(coeffA_part+                   \n"
  "                                   duCoeffB*(downCoeff+upCoeff)); \n"
  "       result=coeffA*(ref+wCoeffB*west+eCoeffB*east               \n"
  "                         +sCoeffB*south+nCoeffB*north             \n"
  "                         +duCoeffB*(downCoeff*down+upCoeff*up));  \n"
  " #if CLIP_NEGATIVE_VALUES                                         \n"
  "       result=fmaxf(0.0f,result);                                 \n"
  " #endif                                                           \n"
  "       }                                                          \n"
  "     to_array[cId]=result;                                        \n"
  "     downCoeff*=DIV_H2_PROGRESSION;                               \n"
  "     upCoeff*=DIV_H2_PROGRESSION;                                 \n"
  "     wId+=XY_COUNT; eId+=XY_COUNT;                                \n"
  "     sId+=XY_COUNT; nId+=XY_COUNT;                                \n"
  "     }                                                            \n"
  "   }                                                              \n"
  " }                                                                \n");
Substrate_buildProgram(&sub->iterPrg,dev,"iterPrg",&src);
#if ENABLE_RENDERING
  CudaBuffer_initFromGl(&sub->gpu_tex,&sub->compStream,true,
                        Rendering_sideTextureBuffer());
  Text_clear(&src);
  Text_appendText(&src,&header);
  Text_appendStr(&src,
    " extern \"C\" __global__                                          \n"
    " void                                                             \n"
    " texPrg(const real32_t * __restrict__ from_array,                 \n"
    "        int * __restrict__ tex_array,                             \n"
    "        const int sideWidth,                                      \n"
    "        const int sideHeight,                                     \n"
    "        const int hSkip,                                          \n"
    "        const int vSkip,                                          \n"
    "        const int pSkip,                                          \n"
    "        const real32_t valueFactor,                               \n"
    "        const real32_t valueIncrement)                            \n"
    " {                                                                \n"
    " const int sz=sideWidth*sideHeight;                               \n"
    " for(int id=GROUP_SIZE*blockIdx.x+threadIdx.x;                    \n"
    "     id<sz;id+=GRID_SIZE)                                         \n"
    "   {                                                              \n"
    "   const int vId=id/sideWidth;                                    \n"
    "   const int hId=id-(vId*sideWidth);                              \n"
#if 1
    "   const int valueId=hId*hSkip+vId*vSkip+pSkip;                   \n"
    "   real32_t value=__fmaf_rn(from_array[valueId],                  \n"
    "                            valueFactor,valueIncrement);          \n"
    "   // [0.5;0.75]-->[0;1]                                          \n"
    "   const real32_t red=__saturatef(4.0f*(value-0.5f));             \n"
    "   // [0;0.25]-->[0;1]  [0.75;1]-->[1;0]                          \n"
    "   const real32_t green=__saturatef(2.0f-4.0f*fabsf(value-0.5f)); \n"
    "   // [0.25;0.5]-->[1;0]                                          \n"
    "   const real32_t blue=__saturatef(1.0f+4.0f*(0.25f-value));      \n"
#else
    "   const real32_t red=hId/real32_t(sideWidth);                    \n"
    "   const real32_t green=vId/real32_t(sideHeight);                 \n"
    "   const real32_t blue=(hId&1)==(vId&1) ? 0.25f : 0.75f;          \n"
#endif
    "   const int texId=vId*sideWidth+hId;                             \n"
    "   // little-endian BGRA                                          \n"
    "   tex_array[texId]=(((unsigned char)(255.0f*blue))<<0)           \n"
    "                   |(((unsigned char)(255.0f*green))<<8)          \n"
    "                   |(((unsigned char)(255.0f*red))<<16)           \n"
    "                   |(255<<24); // alpha                           \n"
    "   }                                                              \n"
    " }                                                                \n");
  Substrate_buildProgram(&sub->texPrg,dev,"texPrg",&src);
  sub->sides[WEST_SIDE]=0;
  sub->sides[EAST_SIDE]=cst_substrate_sampling-1;
  sub->sides[SOUTH_SIDE]=0;
  sub->sides[NORTH_SIDE]=cst_substrate_sampling-1;
  sub->sides[DOWN_SIDE]=cst_substrate_plane_count-1;
  sub->sides[UP_SIDE]=0;
#endif
Text_destroy(&src);
Text_destroy(&header);
}

void
Substrate_reset(void)
{
Substrate * RESTRICT sub=&sd.substrate;
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,substrate_sampling);
IMPORT_INT_CONSTANT(cst,substrate_plane_count);
const real32_t zero=0.0f;
const int xyCount=cst_substrate_sampling*cst_substrate_sampling;
const int xyzCount=xyCount*cst_substrate_plane_count;
int argCount;
const void *args[32];
for(int i=0;i<SUBSTANCE_COUNT;++i)
  {
  Substance * RESTRICT s=&sub->substances[i];
  for(int j=0;j<3;++j)
    {
    argCount=0;
    args[argCount++]=CudaBuffer_programArg(&s->gpu_c[j]);
    args[argCount++]=&xyzCount;
    args[argCount++]=&s->initValue;
    CudaProgram_launch(&sub->rstPrg,
                       sub->groupCount,1,1,
                       sub->groupSize,1,1,
                       0,args,&sub->compStream);
    argCount=0;
    args[argCount++]=CudaBuffer_programArg(&s->gpu_dc[j]);
    args[argCount++]=&xyCount;
    args[argCount++]=&zero;
    CudaProgram_launch(&sub->rstPrg,
                       sub->groupCount,1,1,
                       sub->groupSize,1,1,
                       0,args,&sub->compStream);
    }
  }
CudaStream_hostSync(&sub->compStream);
const int topSize=xyCount*(int)sizeof(real32_t);
for(int i=0;i<SUBSTANCE_COUNT;++i)
  {
  Substance * RESTRICT s=&sub->substances[i];
  for(int j=0;j<3;++j)
    {
    CudaBuffer_receive(&s->gpu_c[j],0,topSize,s->top_c[j],
                       &sub->recvStream);
    }
  }
CudaStream_hostSync(&sub->recvStream);
}

static
void
Substrate_computeDiffusion(Substrate * RESTRICT sub,
                           const int startState,
                           const int deltaState,
                           const int resultState)
{
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,substrate_iteration_count);
IMPORT_INT_CONSTANT(cst,substrate_dt_divisor);
CudaBuffer *diff_array=&sub->gpu_diff;
CudaBuffer *ref_array=&sub->gpu_ref;
int argCount;
const void *args[32];
for(int i=0;i<SUBSTANCE_COUNT;++i)
  {
  Substance * RESTRICT s=&sub->substances[i];
  CudaBuffer *delta_array=&s->gpu_dc[deltaState];
  CudaBuffer *result_arrays[2]={ &sub->gpu_tmp, &s->gpu_c[resultState] };
  // ensures the very last array will be gpu_c[resultState]
  int toId=cst_substrate_dt_divisor&cst_substrate_iteration_count&1;
  CudaBuffer *last_array=&s->gpu_c[startState]; // will start from here
  for(int step=0;step<cst_substrate_dt_divisor;++step)
    {
    CudaBuffer *prev_array=last_array;
    argCount=0;
    args[argCount++]=&s->diffusivity;
    args[argCount++]=CudaBuffer_programArg(diff_array);
    args[argCount++]=CudaBuffer_programArg(prev_array);
    args[argCount++]=CudaBuffer_programArg(delta_array);
    args[argCount++]=CudaBuffer_programArg(ref_array);
    CudaProgram_launch(&sub->refPrg,
                       sub->groupCount,1,1,
                       sub->groupSize,1,1,
                       0,args,&sub->compStream);
    for(int iter=0;iter<cst_substrate_iteration_count;++iter)
      {
      prev_array=last_array;
      last_array=result_arrays[toId];
      argCount=0;
      args[argCount++]=&s->diffusivity;
      args[argCount++]=CudaBuffer_programArg(diff_array);
      args[argCount++]=CudaBuffer_programArg(ref_array);
      args[argCount++]=CudaBuffer_programArg(prev_array);
      args[argCount++]=CudaBuffer_programArg(last_array);
      CudaProgram_launch(&sub->iterPrg,
                         sub->groupCount,1,1,
                         sub->groupSize,1,1,
                         0,args,&sub->compStream);
      toId=1-toId; // alternate next last_array
      }
    }
  }
}

#if SYNCHRONISE_SUBSTRATE

void
Substrate_diffuse(void)
{
Substrate * RESTRICT sub=&sd.substrate;
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,substrate_sampling);
IMPORT_INT_CONSTANT(cst,substrate_plane_count);
const int fromState2=Scheduler_fromState2(&sd.sched);
const int toState2=Scheduler_toState2(&sd.sched);
const int topSize=cst_substrate_sampling*cst_substrate_sampling*
                  (int)sizeof(real32_t);
for(int i=0;i<SUBSTANCE_COUNT;++i)
  {
  Substance * RESTRICT s=&sub->substances[i];
  CudaBuffer_send(&s->gpu_dc[toState2],0,topSize,
                  s->top_dc[toState2],&sub->sendStream);
  }
CudaMarker_set(&sub->sendSync[toState2],&sub->sendStream);
CudaMarker_deviceSync(&sub->sendSync[toState2],&sub->compStream);
if(sd.pause)
  {
  const int topCellCount=cst_substrate_sampling*cst_substrate_sampling;
  const int topSize=topCellCount*(int)sizeof(real32_t);
  const int bufSize=topSize*cst_substrate_plane_count;
  for(int i=0;i<SUBSTANCE_COUNT;++i)
    {
    Substance * RESTRICT s=&sub->substances[i];
    // FIXME: wouldn't it be better to swap buffers on the host side?
    CudaBuffer *tmp_buf=&sub->gpu_tmp;
    CudaBuffer *from_buf=&s->gpu_c[fromState2];
    CudaBuffer *to_buf=&s->gpu_c[toState2];
    CudaBuffer_copy(tmp_buf,to_buf,0,0,bufSize,&sub->compStream);
    CudaBuffer_copy(to_buf,from_buf,0,0,bufSize,&sub->compStream);
    CudaBuffer_copy(from_buf,tmp_buf,0,0,bufSize,&sub->compStream);
    }
  }
else
  {
  Substrate_computeDiffusion(sub,fromState2,toState2,toState2);
  }
CudaMarker_set(&sub->compSync[toState2],&sub->compStream);
CudaMarker_deviceSync(&sub->compSync[toState2],&sub->recvStream);
for(int i=0;i<SUBSTANCE_COUNT;++i)
  {
  Substance * RESTRICT s=&sub->substances[i];
  CudaBuffer_receive(&s->gpu_c[toState2],0,topSize,
                     s->top_c[toState2],&sub->recvStream);
  }
CudaStream_hostSync(&sub->recvStream);
}

#else

void
Substrate_beginDiffusion(void)
{
Substrate * RESTRICT sub=&sd.substrate;
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,substrate_sampling);
IMPORT_INT_CONSTANT(cst,substrate_plane_count);
const int fromState3=Scheduler_fromState3(&sd.sched);
const int currentState3=Scheduler_currentState3(&sd.sched);
const int toState3=Scheduler_toState3(&sd.sched);
CudaMarker_deviceSync(&sub->sendSync[fromState3],&sub->compStream);
if(sd.pause)
  {
  const int topCellCount=cst_substrate_sampling*cst_substrate_sampling;
  const int topSize=topCellCount*(int)sizeof(real32_t);
  const int bufSize=topSize*cst_substrate_plane_count;
  for(int i=0;i<SUBSTANCE_COUNT;++i)
    {
    Substance * RESTRICT s=&sub->substances[i];
    // FIXME: wouldn't it be better to swap buffers on the host side?
    CudaBuffer *tmp_buf=&sub->gpu_tmp;
    CudaBuffer *from_buf=&s->gpu_c[fromState3];
    CudaBuffer *current_buf=&s->gpu_c[currentState3];
    CudaBuffer *to_buf=&s->gpu_c[toState3];
    CudaBuffer_copy(tmp_buf,to_buf,0,0,bufSize,&sub->compStream);
    CudaBuffer_copy(to_buf,current_buf,0,0,bufSize,&sub->compStream);
    CudaBuffer_copy(current_buf,from_buf,0,0,bufSize,&sub->compStream);
    CudaBuffer_copy(from_buf,tmp_buf,0,0,bufSize,&sub->compStream);
    }
  }
else
  {
  Substrate_computeDiffusion(sub,currentState3,fromState3,toState3);
  }
CudaMarker_set(&sub->compSync[toState3],&sub->compStream);
CudaMarker_deviceSync(&sub->compSync[toState3],&sub->recvStream);
const int topSize=cst_substrate_sampling*cst_substrate_sampling*
                  (int)sizeof(real32_t);
for(int i=0;i<SUBSTANCE_COUNT;++i)
  {
  Substance * RESTRICT s=&sub->substances[i];
  CudaBuffer_receive(&s->gpu_c[toState3],0,topSize,
                     s->top_c[toState3],&sub->recvStream);
  }
CudaMarker_set(&sub->recvSync[toState3],&sub->recvStream);
}

void
Substrate_endDiffusion(void)
{
Substrate * RESTRICT sub=&sd.substrate;
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,substrate_sampling);
const int currentState3=Scheduler_currentState3(&sd.sched);
const int toState3=Scheduler_toState3(&sd.sched);
const int topSize=cst_substrate_sampling*cst_substrate_sampling*
                  (int)sizeof(real32_t);
for(int i=0;i<SUBSTANCE_COUNT;++i)
  {
  Substance * RESTRICT s=&sub->substances[i];
  CudaBuffer_send(&s->gpu_dc[toState3],0,topSize,
                  s->top_dc[toState3],&sub->sendStream);
  }
CudaMarker_set(&sub->sendSync[toState3],&sub->sendStream);
CudaMarker_hostSync(&sub->recvSync[currentState3]); // future fromState3
}

#endif

#if ENABLE_RENDERING

void
Substrate_updateTextureBuffer(int side,
                              int sideWidth,
                              int sideHeight,
                              int rendered)
{
Substrate * RESTRICT sub=&sd.substrate;
const ConstantRecord * RESTRICT cst=ConstantRecord_get();
IMPORT_INT_CONSTANT(cst,substrate_sampling);
CudaBuffer *values;
real32_t minValue,maxValue;
bool onlyTop;
switch(rendered)
  {
  case SUBSTANCE_COUNT+0:
    {
    IMPORT_R32_CONSTANT(cst,diffusivity_plot_amount);
    IMPORT_R32_CONSTANT(cst,diffusivity_plot_value);
    values=&sub->gpu_diff;
    minValue=0.0f;
    maxValue=cst_diffusivity_plot_amount
             ? fmaxf(1.0f,cst_diffusivity_plot_value)
             : 2.0f;
    onlyTop=true;
    break;
    }
  default:
    {
    Substance * RESTRICT s=&sub->substances[rendered];
#if SYNCHRONISE_SUBSTRATE
    const int lastState=Scheduler_fromState2(&sd.sched);
#else
    const int lastState=Scheduler_currentState3(&sd.sched);
#endif
    values=&s->gpu_c[lastState];
    minValue=s->minDisplay;
    maxValue=s->maxDisplay;
    onlyTop=false;
    break;
    }
  }
const int xyCount=cst_substrate_sampling*cst_substrate_sampling;
int hSkip=-1,vSkip=-1,pSkip=-1;
switch(side)
  {
  case WEST_SIDE:
  case EAST_SIDE:
    {
    hSkip=cst_substrate_sampling; // y
    vSkip=onlyTop ? 0 : xyCount; // z
    pSkip=1; // x
    break;
    }
  case SOUTH_SIDE:
  case NORTH_SIDE:
    {
    hSkip=1; // x
    vSkip=onlyTop ? 0 : xyCount; // z
    pSkip=cst_substrate_sampling; // y
    break;
    }
  case DOWN_SIDE:
  case UP_SIDE:
    {
    hSkip=1; // x
    vSkip=cst_substrate_sampling; // y
    pSkip=onlyTop ? 0 : xyCount; // z
    break;
    }
  }
pSkip*=sub->sides[side];
const real32_t valueFactor=1.0f/(maxValue-minValue);
const real32_t valueIncrement=-minValue*valueFactor;
CudaBuffer_acquireGl(&sub->gpu_tex,&sub->compStream);
int argCount=0;
const void *args[32];
args[argCount++]=CudaBuffer_programArg(values);
args[argCount++]=CudaBuffer_programArg(&sub->gpu_tex);
args[argCount++]=&sideWidth;
args[argCount++]=&sideHeight;
args[argCount++]=&hSkip;
args[argCount++]=&vSkip;
args[argCount++]=&pSkip;
args[argCount++]=&valueFactor;
args[argCount++]=&valueIncrement;
CudaProgram_launch(&sub->texPrg,
                   sub->groupCount,1,1,
                   sub->groupSize,1,1,
                   0,args,&sub->compStream);
CudaBuffer_releaseGl(&sub->gpu_tex,&sub->compStream);
}

#endif

//----------------------------------------------------------------------------
